/*
 * gMix open source project - https://svs.informatik.uni-hamburg.de/gmix/
 * Copyright (C) 2012  Karl-Peter Fuchs
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package evaluation.simulator.statistics;

import gnu.trove.TDoubleArrayList;
import gnu.trove.TDoubleProcedure;

import java.math.BigDecimal;

import evaluation.simulator.core.Event;
import evaluation.simulator.core.EventExecutor;
import evaluation.simulator.core.Simulator;
import evaluation.simulator.networkComponent.Identifiable;


public class Statistics implements EventExecutor {

	private static final int INITIAL_SIZE = 10000;
	private static GetSumHelper getSumHelper = new GetSumHelper();
	private static DisplayAllHelper displayAllHelper = new DisplayAllHelper(false);
	private static DisplayAllHelper displayAllSeparatedHelper = new DisplayAllHelper(true);
	private static GetPercentageHelper getPercentageHelper = new GetPercentageHelper();
	
	private boolean isCumulativeStatisticsObject = false; // statistics object used to calculate EvaluationTypes (e.g. average or minimum) from recorded statistics 
	
	private TDoubleArrayList[] recordedValues;
	private BigDecimal[] summedUpValues;
	//private Identifiable owner;
	private static boolean recordStatistics = false;
	private static Simulator simulator;
	
	
	public Statistics(Identifiable owner) { // TODO: wird "owner" noch für irgendetwas benötigt?
		//this.owner = owner;
		GeneralStatistics.registerStatisticsObject(this);
		recordedValues = new TDoubleArrayList[StatisticsType.values().length];
		summedUpValues = new BigDecimal[StatisticsType.values().length];
	}
	
	
	// constructor for class GeneralStatistics (selbstreferenzen vermeiden...)
	protected Statistics() {
		isCumulativeStatisticsObject = true;
		recordedValues = new TDoubleArrayList[StatisticsType.values().length];
		summedUpValues = new BigDecimal[StatisticsType.values().length];
	}
	
	
	private Statistics(int i) {}
	
	
	public static void setSimulator(Simulator simulator) {
		Statistics.simulator = simulator;
		simulator.ts_recordStatisticsStart = Simulator.settings.getPropertyAsInt("RECORD_STATISTICS_FROM_IN_MS");
		Statistics s = new Statistics(0);
		// set start
		if (simulator.ts_recordStatisticsStart == 0) {
			recordStatistics = true;
		} else {
			simulator.scheduleEvent(new Event(s, simulator.ts_recordStatisticsStart, StatisticsEvent.START_RECORDING), s);
		}
		// set fixed end (if specified)
		if (Simulator.settings.getProperty("SIMULATION_END").equals("SIMULATION_TIME_END")) {
			int recordStatisticsEnd = Simulator.settings.getPropertyAsInt("SIMULATION_TIME_LIMIT_IN_MS");
			simulator.scheduleEvent(new Event(s, recordStatisticsEnd, StatisticsEvent.STOP_RECORDING), s);
		}
	}
	
	
	public void addValue(double value, StatisticsType statisticsType) {
		
		if (!statisticsType.isActivated || (!recordStatistics && !isCumulativeStatisticsObject))
			return;
		
		if (statisticsType.intendedEvaluations.length == 1 && (statisticsType.intendedEvaluations[0] == EvaluationType.SUM || statisticsType.intendedEvaluations[0] == EvaluationType.EVENTS_PER_SECOND || statisticsType.intendedEvaluations[0] == EvaluationType.VOLUME_PER_SECOND || statisticsType.intendedEvaluations[0] == EvaluationType.EVENTS_PER_SECOND_AND_CLIENT)) {
			
			if (summedUpValues[statisticsType.ordinal()] == null)
				summedUpValues[statisticsType.ordinal()] = new BigDecimal(0);
			summedUpValues[statisticsType.ordinal()] = summedUpValues[statisticsType.ordinal()].add(new BigDecimal(value));

		} else {

			if (recordedValues[statisticsType.ordinal()] == null)
				recordedValues[statisticsType.ordinal()] = new TDoubleArrayList(INITIAL_SIZE);
			recordedValues[statisticsType.ordinal()].add(value);
			
		}
		
	}

	
	public boolean containsResultsFor(StatisticsType statisticsType) {
		
		if (statisticsType.intendedEvaluations.length == 1 && (statisticsType.intendedEvaluations[0] == EvaluationType.SUM || statisticsType.intendedEvaluations[0] == EvaluationType.EVENTS_PER_SECOND || statisticsType.intendedEvaluations[0] == EvaluationType.EVENTS_PER_SECOND || statisticsType.intendedEvaluations[0] == EvaluationType.VOLUME_PER_SECOND || statisticsType.intendedEvaluations[0] == EvaluationType.EVENTS_PER_SECOND_AND_CLIENT)) 
			return summedUpValues[statisticsType.ordinal()] != null;
		else
			return recordedValues[statisticsType.ordinal()] != null;
		
	}
	
	
	public double getResultDouble(	StatisticsType statisticsType, 
									EvaluationType evaluationType
									) throws NullPointerException {
		
		
		if (evaluationType.dataType != EvaluationType.DATA_TYPE_DOUBLE)
			throw new RuntimeException("ERROR: wrong data type!");
		
		try {
			
			switch (evaluationType) {
			
				case MIN:
					return recordedValues[statisticsType.ordinal()].min();
		
				case MAX:
					return recordedValues[statisticsType.ordinal()].max();
			
				case AVG:
					BigDecimal sum = getSumHelper.getSum(recordedValues[statisticsType.ordinal()]);
					BigDecimal avg = sum.divide(new BigDecimal(recordedValues[statisticsType.ordinal()].size()), 5, BigDecimal.ROUND_HALF_UP);
					return avg.doubleValue();
				
				case MEDIAN:
					recordedValues[statisticsType.ordinal()].sort();
					int mid = (int)Math.round((double)recordedValues[statisticsType.ordinal()].size() / 2d);
					if (mid == 0)
						throw new RuntimeException("ERROR: no data!");
					else
						return recordedValues[statisticsType.ordinal()].get(mid-1);
				
				case PERCENTAGE:
					if (isCumulativeStatisticsObject) {
						sum = getSumHelper.getSum(recordedValues[statisticsType.ordinal()]);
						avg = sum.divide(new BigDecimal(recordedValues[statisticsType.ordinal()].size()), 5, BigDecimal.ROUND_HALF_UP);
						return avg.doubleValue();
					} else
						return getPercentageHelper.getPercentage(recordedValues[statisticsType.ordinal()]);
					
				case VALUE_LIST:
					if (isCumulativeStatisticsObject) {
						System.out.println(statisticsType); 
						return 0d; // dummy
					} else {
						displayAllHelper.displayAll(recordedValues[statisticsType.ordinal()]);
						return 0d; // dummy
					}
					
				case SEPARATED_VALUE_LIST:
					if (isCumulativeStatisticsObject) {
						System.out.println(statisticsType); 
						return 0d; // dummy
					} else {
						displayAllSeparatedHelper.displayAll(recordedValues[statisticsType.ordinal()]);
						return 0d; // dummy
					}
					
				default:
					throw new RuntimeException("ERROR: unknown or not allowed evaluation type!");
				
			}
			
		} catch (IllegalStateException e) {
			
			e.printStackTrace();
			throw new RuntimeException("ERROR: no data!");
			
		}
		
		
	}

	private boolean displayPerClient = true; // TODO
	
	public BigDecimal getResultBigDecimal(	StatisticsType statisticsType, 
											EvaluationType evaluationType 
											) throws NullPointerException {
		
		if (evaluationType.dataType != EvaluationType.DATA_TYPE_COUNTER)
			throw new RuntimeException("ERROR: wrong data type!");
		
		try {
			
			switch (evaluationType) {
			
				case SUM:
					return summedUpValues[statisticsType.ordinal()];

				case EVENTS_PER_SECOND:
					if (isCumulativeStatisticsObject) {
						BigDecimal numberOfEvents = summedUpValues[statisticsType.ordinal()];
						BigDecimal simulationTime = new BigDecimal(simulator.ts_recordStatisticsEnd - simulator.ts_recordStatisticsStart).divide(new BigDecimal(1000), 5, BigDecimal.ROUND_HALF_UP); // to sec;
						return numberOfEvents.divide(simulationTime, 5, BigDecimal.ROUND_HALF_UP);
					} else 
						return summedUpValues[statisticsType.ordinal()];
				
				case EVENTS_PER_SECOND_AND_CLIENT:
					if (isCumulativeStatisticsObject) {
						BigDecimal numberOfEvents = summedUpValues[statisticsType.ordinal()];
						BigDecimal simulationTime = new BigDecimal(simulator.ts_recordStatisticsEnd - simulator.ts_recordStatisticsStart).divide(new BigDecimal(1000), 5, BigDecimal.ROUND_HALF_UP); // to sec;
						BigDecimal eventsPerSecond = numberOfEvents.divide(simulationTime, 5, BigDecimal.ROUND_HALF_UP);
						return eventsPerSecond.divide(new BigDecimal(simulator.getClients().size()), 5, BigDecimal.ROUND_HALF_UP); // divide number of clients
					} else 
						return summedUpValues[statisticsType.ordinal()];
					
				case VOLUME_PER_SECOND:
					if (isCumulativeStatisticsObject) {
						BigDecimal byteCounter = summedUpValues[statisticsType.ordinal()].divide(new BigDecimal(1024), 5, BigDecimal.ROUND_HALF_UP); // to kbyte
						System.out.println("byteCounter (kbyte): " +byteCounter); 
						BigDecimal simulationTime = new BigDecimal(simulator.ts_recordStatisticsEnd - simulator.ts_recordStatisticsStart).divide(new BigDecimal(1000), 5, BigDecimal.ROUND_HALF_UP); // to sec;
						return byteCounter.divide(simulationTime, 5, BigDecimal.ROUND_HALF_UP);
					} else {
						if (displayPerClient) {
							BigDecimal byteCounter = summedUpValues[statisticsType.ordinal()].divide(new BigDecimal(1024), 5, BigDecimal.ROUND_HALF_UP); // to kbyte
							BigDecimal simulationTime = new BigDecimal(simulator.ts_recordStatisticsEnd - simulator.ts_recordStatisticsStart).divide(new BigDecimal(1000), 5, BigDecimal.ROUND_HALF_UP); // to sec;
							BigDecimal result = byteCounter.divide(simulationTime, 5, BigDecimal.ROUND_HALF_UP);
							System.out.println(result); 
						}
						return summedUpValues[statisticsType.ordinal()];
					}
					
				default:
					throw new RuntimeException("ERROR: unknown evaluation type!");
				
			}
			
		} catch (IllegalStateException e) {
			
			e.printStackTrace();
			throw new RuntimeException("ERROR: no data!");
			
		}
		
	}
	
	
	static class GetPercentageHelper implements TDoubleProcedure {
		
		double sumPos = 0;
		double sumNeg = 0;
		

		public double getPercentage(TDoubleArrayList data) {
			data.forEach(getPercentageHelper);
			double result = sumNeg/(sumPos+sumNeg);
			//System.out.println("sumNeg: " +sumNeg); 
			//System.out.println("sumPos: " +sumPos); 
			//double simulationTime = simulator.ts_recordStatisticsEnd - simulator.ts_recordStatisticsStart;
			//double msgPerSec = (sumPos+sumNeg) / (simulationTime/1000d);
			//double msgDropPerSec = sumNeg / (simulationTime/1000d);
			//System.out.println("sent in total: " +(sumPos+sumNeg) +" messages in " +simulationTime +"ms -> " +msgPerSec +" msg/sek (rate: " +1d/((simulationTime/1000d)/(sumPos+sumNeg)) +")" ); 
			//System.out.println("getPercentage-res: " +result*100d +"%"); // TODO: remove
			//System.out.println("dropped in total: " +sumNeg +" messages in " +simulationTime +"ms -> " +msgDropPerSec +" msg/sek (rate: " +1d/((simulationTime/1000d)/(sumNeg)) +")" ); 
			sumNeg = 0;
			sumPos = 0;
			return result;
		}
		
		
		@Override
		public boolean execute(double val) {
			if (val == 1.0d)
				sumPos++;
			else if (val == -1.0d)
				sumNeg++;
			else
				throw new RuntimeException("use only 1.0d and -1.0d when recording statistics for PERCENTAGE EvaluationType"); 
			return true;
		}
		
	}
	
	
	static class GetSumHelper implements TDoubleProcedure {
		
		BigDecimal sum = new BigDecimal(0);
		
		
		public BigDecimal getSum(TDoubleArrayList data) {
			data.forEach(getSumHelper);
			BigDecimal result = sum;
			sum = new BigDecimal("0");
			return result;
		}
		
		
		@Override
		public boolean execute(double val) {
			sum = sum.add(new BigDecimal(""+val));
			return true;
		}
		
	}

	
	static class DisplayAllHelper implements TDoubleProcedure {
		
		private boolean isSeparated;
		
		public DisplayAllHelper(boolean isSeparated) {
			this.isSeparated = isSeparated;
		}
		
		
		public void displayAll(TDoubleArrayList data) {
			data.forEach(displayAllHelper);
			if (isSeparated)
				System.out.println(); 
		}
		
		
		@Override
		public boolean execute(double val) {
			System.out.println(val); 
			return true;
		}
		
	}
	

	@Override
	public void executeEvent(Event event) {
		
		if (event.getEventType() == StatisticsEvent.START_RECORDING) {
			recordStatistics = true;
			System.out.println("### START recording statistics"); 
		} else if (event.getEventType() == StatisticsEvent.STOP_RECORDING) {
			System.out.println("### STOP recording statistics"); 
			recordStatistics = false; 
			simulator.stopSimulation("simulation-time limit reached (variable SIMULATION_TIME_LIMIT_IN_MS in experiment config)");
		} else {
			throw new RuntimeException("ERROR! received unsupported event!" +event);
		}
		
	}
	
}
